/*
 *	xspray
 */

static char *rcsid="$Id: xspray.c,v 1.1 1999-09-20 21:23:17+09 kabe Exp $";

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/Xresource.h>
#include <stdio.h>
#include <stdlib.h>	/*drand48*/
#include <math.h>
#include <strings.h>	/*strcmp*/

XrmDatabase	gDatabase;
XrmOptionDescRec	gOptable[] = {
{"-display",	"*display",	XrmoptionSepArg,	(XPointer)NULL },
{"-fg",	"*Foreground",	XrmoptionSepArg,	(XPointer)NULL },
{"-bg",	"*Background",	XrmoptionSepArg,	(XPointer)NULL },
{"-dump",	"*dumpTrace",	XrmoptionNoArg,	(XPointer)"True"},
{"-trace",	"*traceFile",	XrmoptionSepArg,	(XPointer)NULL },
{"-noquit",	"*noquit",	XrmoptionNoArg,	(XPointer)"True" },
{"-radius",	"*radius",	XrmoptionSepArg,	(XPointer)NULL },
{"-accel",	"*dragAccel",	XrmoptionSepArg,	(XPointer)NULL },
};
const int	gOptableN = (sizeof(gOptable)/sizeof(*gOptable));

int	sprayRadius = 30;
double	dragAccel = 0.003;

Bool	dumpTrace = False;
char	*traceFile = NULL;
Bool	continueAfterTrace = False;

static int speckle();
static int drag();
Cursor CreateSpraycanCursor();
char	*getStringResource();
Bool getBoolResource();

int
main(argc, argv)
int	argc;
char	*argv[];
{
	Display	*d;
	Window	w,root;
	GC	gc1,gc3, *curGC=NULL;
	Bool	finish = False;
	Cursor	spraycan;

	srand48(time(NULL));

	/* parse command line */
	XrmInitialize();
	XrmParseCommand(&gDatabase, gOptable, gOptableN, "xspray", &argc, argv);

	/* retrieve globals */
	dumpTrace = getBoolResource(gDatabase, "xspray.dumpTrace","Xspray.Debug", False);
	traceFile = getStringResource(gDatabase, "xspray.traceFile","Xspray.Filename");
	continueAfterTrace = getBoolResource(gDatabase, "xspray.noquit","Xspray.Debug", False);
	{char *p;
	if (p=getStringResource(gDatabase, "xspray.radius","Xspray.Radius"))
		sprayRadius = atoi(p);
	if (p=getStringResource(gDatabase, "xspray.dragAccel","Xspray.dragAccel"))
		dragAccel = atof(p);

	}

	d = XOpenDisplay(getStringResource(gDatabase, "xspray.display","Xspray.Display"));
	if (!d) {
		fprintf(stderr,"Cannot open Display.\n"); exit(1);
	}
	root = DefaultRootWindow(d);

	/* create a input window of size of root */
	{XSetWindowAttributes attr; unsigned long mask = 0;
	unsigned int rwidth, rheight, dummy; Window dw;
	XGetGeometry(d,root, &dw, &dummy,&dummy, &rwidth,&rheight, &dummy,&dummy);
	attr.override_redirect = True;	mask |= CWOverrideRedirect;
	w = XCreateWindow(d,root, 0,0, rwidth,rheight, 0,
		0,InputOnly,CopyFromParent, mask,&attr
	);
	}

	/* standard GC */
	{XGCValues	gval;
	unsigned long	mask = 0;
	XColor	cols,colx;
	char	*cname;

	gval.subwindow_mode = IncludeInferiors; mask |= GCSubwindowMode;
	gval.function = GXcopy;	mask |= GCFunction;

	gval.foreground = WhitePixel(d, DefaultScreen(d)); mask |= GCForeground;
	if ((cname=getStringResource(gDatabase, "xspray.foreground","Xspray.Foreground")) && 
	    XAllocNamedColor(d, DefaultColormap(d,DefaultScreen(d)), 
		cname, &cols, &colx
	))	gval.foreground = cols.pixel;
	gc1 = XCreateGC(d,root, mask,&gval);

	gval.foreground = BlackPixel(d, DefaultScreen(d));
	if ((cname = getStringResource(gDatabase, "xspray.background","Xspray.Background")) &&
	    XAllocNamedColor(d, DefaultColormap(d,DefaultScreen(d)), 
		cname, &cols, &colx
	))	gval.foreground = cols.pixel;
	gc3 = XCreateGC(d,root, mask,&gval);
	}

	spraycan = CreateSpraycanCursor(d);

	(!traceFile || continueAfterTrace) && fprintf(stderr, "Press Button2 to exit...\n");
	dumpTrace && printf("# xspray trace file\n");

	XSelectInput(d, w, ButtonPressMask | ButtonReleaseMask | ButtonMotionMask |
		StructureNotifyMask |
		KeyPressMask
	);
	XMapWindow(d, w);
	XDefineCursor(d, w, spraycan);

	while(!finish) {
		static XEvent	event;
		static int	hx,hy;	/* cursor position */
		static double	dragx,dragy;	/* spraying center; follows cursor position */
		static Bool	spraying = False;
		static int	scount;

		if (XEventsQueued(d,QueuedAfterFlush) || !spraying) {
			XNextEvent(d, &event);
			switch(event.type) {
			  case ButtonPress:
				switch (event.xbutton.button) {
				  case Button1:	curGC= &gc1; break;
				  case Button3: curGC= &gc3; break;
				  default: goto no1;
				}
				spraying = True;
				dragx = hx = event.xbutton.x;
				dragy = hy = event.xbutton.y;
				dumpTrace && printf("S %d %d\n", hx,hy);
				scount=0;
			    no1: ;
				break;
			  case ButtonRelease:
				/* catch quit button here
				   to eliminate "paste" on xterm */
				if (event.xbutton.button == Button2) {
					finish = True; break;
				}
				spraying && dumpTrace && printf("M %d %d %d\n", hx,hy,scount);
				spraying = False;
				break;
			  case MotionNotify:
				if (spraying) {
					hx = event.xmotion.x;
					hy = event.xmotion.y;
					dumpTrace && printf("M %d %d %d\n", hx,hy,scount);
					scount=0;
				}
				break;
			  case KeyPress: {
				KeySym ksym = XLookupKeysym(&event.xkey,0);
				if (ksym == XK_q) {
					finish = True; break;
				}
				}
				break;
			  case MapNotify: {
				XWindowAttributes attr;
				if (traceFile) {
					sprayAFile(d,root,gc1,traceFile);
					if (!continueAfterTrace) finish=True;
				}
				XGetWindowAttributes(d,w,&attr);
				XSelectInput(d,w, attr.your_event_mask & ~StructureNotifyMask);
				}
				break;
			}
		} else {
			/* no events */
			if (spraying) {
				speckle(d,root,*curGC, (int)dragx,(int)dragy, sprayRadius);
				drag(&dragx,&dragy,dragAccel, hx,hy);
scount++;
			}
		}
	}
}

static int
scatter( radius, power, xret,yret )
int	radius;
int	power;	/*notused*/
int	*xret, *yret;
{
	double	r1, th;
	r1 = (1.0 - sqrt(drand48())) * radius;	/* triangular distribution */
	th = drand48() * 2 * M_PI;
	*xret = r1 * cos(th);
	*yret = r1 * sin(th);
	return 1;
}

/*
 * single iteration of the spray particle
 */
static int
speckle(d,w,gc,x,y,radius)
Display	*d;
Window	w;
GC	gc;
int	x,y;
int	radius;
{
	int	sx,sy;
	if (scatter( radius, 0, &sx,&sy)) {
		XDrawPoint(d,w,gc, x+sx, y+sy);
	}
}
/*
 * (drag{x,y}) ->accel by ratio ->(home{x,y})
 */
static int
drag(dragx,dragy, ratio, homex,homey)
double	*dragx,*dragy;
double	ratio;
int	homex,homey;
{
	*dragx = (*dragx)*(1.0-ratio) + homex*ratio;
	*dragy = (*dragy)*(1.0-ratio) + homey*ratio;
}

/*
 * some doodles to do beforehand
 */
int
sprayAFile(d,w,gc,filename)
Display	*d;
Window	w;
GC	gc;
char	*filename;
{
	int	i;
	double dx,dy;
	int	hx,hy,scount;
	FILE	*f;
	char	buf[256];

	if (!strcmp(filename, "-")) {
		f = stdin;
	} else {
		f = fopen(filename,"r");
	}
	if (!f) {
		fprintf(stderr,"%s: cannot open\n", filename);
		return 0;
	}
	while(fgets(buf,sizeof(buf),f)) {
		switch (buf[0]) {
		  case '#':	break;
		  case 'S':
			/*	S startx starty	*/
			sscanf(buf, "%*s %d %d", &hx,&hy);
			dx=hx; dy=hy;
			break;
		  case 'M':
			/*	M movex movey speckles	*/
			sscanf(buf, "%*s %d %d %d", &hx,&hy, &scount);
			for(; scount; scount--){
				speckle(d,w,gc, (int)dx,(int)dy, sprayRadius);
				drag(&dx,&dy, dragAccel, hx,hy);
			}
			break;
		}
	}
	fclose(f);
}

Cursor
CreateSpraycanCursor(d)
Display	*d;
{
	XColor	fore,back;
	Pixmap	pcursor,pmask;
#include "spraycan.bmp"
#include "spraycanmask.bmp"

	fore.pixel = BlackPixel(d, DefaultScreen(d));
	XQueryColor(d, DefaultColormap(d, DefaultScreen(d)), &fore);
	back.pixel = WhitePixel(d, DefaultScreen(d));
	XQueryColor(d, DefaultColormap(d, DefaultScreen(d)), &back);

	pcursor = XCreatePixmapFromBitmapData(d,DefaultRootWindow(d),
		spraycan_bits, spraycan_width,spraycan_height,
		1,0, 1
	);
	pmask = XCreatePixmapFromBitmapData(d,DefaultRootWindow(d),
		spraycanmask_bits, spraycanmask_width,spraycanmask_height,
		1,0, 1
	);

	return XCreatePixmapCursor(d, pcursor,pmask, &fore,&back, spraycan_x_hot,spraycan_y_hot);
}


char	*
getStringResource(db, name, class)
XrmDatabase	db;
char	*name, *class;
{
	XrmValue	val;
	char	*type;
	if (XrmGetResource(db, name,class, &type,&val) &&
	    !strcmp(type,"String")
	) {
		return val.addr;
	}
	return NULL;
}

Bool
getBoolResource(db, name,class, defl)
XrmDatabase	db;
char	*name, *class;
Bool	defl;
{
	XrmValue	val;
	char	*type;
	if (XrmGetResource(db, name,class, &type,&val) &&
	    !strcmp(type,"String")
	) {
		if (!strcasecmp(val.addr,"True") || !strcasecmp(val.addr,"Yes"))
			return True;
		return False;
	}
	return defl;
}
